# -*- coding: utf-8 -*-

"""

smallparts.text.transcode

Universal text decoding and encoding functions,
with additional functions to read and write text files.

"""


import codecs
import os.path
import shutil

from smallparts import constants

from smallparts.text import split


# Encodings with byte order marks

BOM_ASSIGNMENTS = (
    (codecs.BOM_UTF32_BE, 'utf_32_be'),
    (codecs.BOM_UTF32_LE, 'utf_32_le'),
    (codecs.BOM_UTF16_BE, 'utf_16_be'),
    (codecs.BOM_UTF16_LE, 'utf_16_le'),
    (codecs.BOM_UTF8, 'utf_8_sig'))

DEFAULT_TARGET_ENCODING = constants.UTF_8
DEFAULT_FALLBACK_ENCODING = constants.CP1252
DEFAULT_LINE_ENDING = constants.LF

SUPPORTED_OUTPUT_LINE_ENDINGS = (
    constants.LF,
    constants.CRLF)

#
# Functions
#


def to_unicode_and_encoding_name(
        bytestring,
        from_encoding=None,
        fallback_encoding=DEFAULT_FALLBACK_ENCODING):
    """Try to decode bytestring to a unicode string
    and return a tuple containing the conversion result
    and the source encoding name.

    If bytestring is not a byte string, a TypeError is raised.

    Otherwise, the following algorithm is used:
        - If an explicit input codec was given, decode it using that codec.
        - Else, try each of the known encodings which use a Byte Order Mark,
          defined in the global BOM_ASSIGNMENTS list.
          If none of these Byte Order Marks was found, try to decode it
          using UTF-8. If that fails, use the fallback codec which is defined
          in the global DEFAULT_FALLBACK_ENCODING variable but can be
          overridden using the parameter fallback_encoding.
    """
    if isinstance(bytestring, (bytes, bytearray)):
        if from_encoding:
            return (bytestring.decode(from_encoding),
                    from_encoding)
        #
        for (bom, encoding) in BOM_ASSIGNMENTS:
            if bytestring.startswith(bom):
                return (bytestring[len(bom):].decode(encoding),
                        encoding)
            #
        #
        try:
            return (bytestring.decode(constants.UTF_8),
                    constants.UTF_8)
        except UnicodeDecodeError:
            return (bytestring.decode(fallback_encoding),
                    fallback_encoding)
        #
    #
    raise TypeError('This function requires bytes or bytearray as input,'
                    ' not {0}.'.format(bytestring.__class__.__name__))


def to_unicode(bytestring,
               from_encoding=None,
               fallback_encoding=DEFAULT_FALLBACK_ENCODING):
    """Wrap to_unicode_and_encoding_name(),
    but return the conversion result only."""
    return to_unicode_and_encoding_name(
        bytestring,
        from_encoding=from_encoding,
        fallback_encoding=fallback_encoding)[0]


def anything_to_unicode(
        input_object,
        from_encoding=None,
        fallback_encoding=DEFAULT_FALLBACK_ENCODING):
    """Safe wrapper around to_unicode() returning the string conversion
    of the input object if it was not a byte string
    """
    try:
        return to_unicode(
            input_object,
            from_encoding=from_encoding,
            fallback_encoding=fallback_encoding)
    except TypeError:
        return str(input_object)
    #


def to_bytes(
        unicode_text,
        to_encoding=DEFAULT_TARGET_ENCODING):
    """Encode unicode_text to a bytes representation
    using the provided encoding
    """
    if isinstance(unicode_text, str):
        return unicode_text.encode(to_encoding)
    #
    raise TypeError('This function requires a unicode string as input,'
                    ' not {0}.'.format(unicode_text.__class__.__name__))


def anything_to_bytes(
        input_object,
        to_encoding=DEFAULT_TARGET_ENCODING,
        from_encoding=None,
        fallback_encoding=DEFAULT_FALLBACK_ENCODING):
    """Encode any given object to a bytes representation
    using the provided encoding, after decoding it to unicode
    using this modules's anything_to_unicode() function
    """
    try:
        return to_bytes(input_object, to_encoding=to_encoding)
    except TypeError:
        return anything_to_unicode(
            input_object,
            from_encoding=from_encoding,
            fallback_encoding=fallback_encoding).encode(to_encoding)
    #


def to_utf8(unicode_text):
    """Encode unicode_text to UTF-8
    using this modules's to_bytes() function
    """
    return to_bytes(unicode_text, to_encoding=constants.UTF_8)


def anything_to_utf8(
        input_object,
        from_encoding=None,
        fallback_encoding=DEFAULT_FALLBACK_ENCODING):
    """Encode any given object to its UTF-8 representation
    using this modules's to_bytes() function
    """
    return anything_to_bytes(input_object,
                             to_encoding=constants.UTF_8,
                             from_encoding=from_encoding,
                             fallback_encoding=fallback_encoding)


def fix_double_utf8_transformation(unicode_text,
                                   wrong_encoding=constants.CP1252):
    """Fix duplicate UTF-8 transformation,
    which is a frequent result of reading UTF-8 encoded text as Latin encoded
    (CP-1252, ISO-8859-1 or similar), resulting in characters like Ã¤Ã¶Ã¼.
    This function reverts the effect.
    """
    if wrong_encoding == constants.UTF_8:
        raise ValueError('This cannot have any effect!')
    #
    return to_unicode(
        to_bytes(
            unicode_text,
            to_encoding=wrong_encoding),
        from_encoding=constants.UTF_8)


def read_from_file(input_file_or_name,
                   from_encoding=None,
                   fallback_encoding=DEFAULT_FALLBACK_ENCODING):
    """Read input file and return its contents as unicode"""
    try:
        file_contents = input_file_or_name.read()
    except AttributeError:
        with open(input_file_or_name,
                  mode=constants.MODE_READ_BINARY) as real_input_file:
            file_contents = real_input_file.read()
        #
    #
    return to_unicode(
        file_contents,
        from_encoding=from_encoding,
        fallback_encoding=fallback_encoding)


def prepare_file_output(unicode_content,
                        to_encoding=DEFAULT_TARGET_ENCODING,
                        line_ending=DEFAULT_LINE_ENDING):
    """Return a bytes representation of unicode_content,
    with the provided line_ending,
    suitable for writing to a file using mode MODE_WRITE_BINARY.
    """
    if line_ending not in SUPPORTED_OUTPUT_LINE_ENDINGS:
        raise ValueError(
            'line_ending must be one of {0!r}!'.format(
                SUPPORTED_OUTPUT_LINE_ENDINGS))
    #
    lines_list = []
    if isinstance(unicode_content, str):
        lines_list.extend(
            split.lines_for_reconstruction(unicode_content))
    else:
        for text_block in unicode_content:
            text_block_lines = split.lines_for_reconstruction(text_block)
            if not text_block_lines:
                lines_list.append('')
                continue
            #
            lines_list.extend(text_block_lines)
        #
    #
    return to_bytes(
        line_ending.join(lines_list),
        to_encoding=to_encoding)


def transcode_file(file_name,
                   to_encoding=DEFAULT_TARGET_ENCODING,
                   from_encoding=None,
                   fallback_encoding=DEFAULT_FALLBACK_ENCODING,
                   line_ending=None):
    """Read the input file and transcode it to the specified encoding in place.
    Raise a ValueError if the detected encoding is the same as the
    target encoding.
    Preserve original line endings except when specified explicitly.
    Rename the original file to the original name with the detected
    encoding name attached.
    """
    with open(file_name,
              mode=constants.MODE_READ_BINARY) as input_file:
        bytes_content = input_file.read()
    #
    unicode_content, detected_encoding = to_unicode_and_encoding_name(
        bytes_content,
        from_encoding=from_encoding,
        fallback_encoding=fallback_encoding)
    #
    # Check content encoding and raise a ValueError
    # if the original contents were already encoded
    # as to_encoding
    if bytes_content == to_bytes(unicode_content, to_encoding=to_encoding):
        raise ValueError(
            'File {0!r} is already encoded in {1!r}!'.format(
                file_name,
                to_encoding))
    #
    # Change line endings if requested
    if line_ending in SUPPORTED_OUTPUT_LINE_ENDINGS:
        unicode_content = line_ending.join(
            split.lines_for_reconstruction(unicode_content))
    elif line_ending is not None:
        raise ValueError(
            'line_ending – if provided – must be one of {0!r}!'.format(
                SUPPORTED_OUTPUT_LINE_ENDINGS))
    #
    # rename the original file to the backup file name
    file_name_root, file_extension = os.path.splitext(file_name)
    backup_file_name = '{0}.{1}{2}'.format(
        file_name_root, detected_encoding, file_extension)
    shutil.move(file_name, backup_file_name)
    #
    with open(file_name,
              mode=constants.MODE_WRITE_BINARY) as output_file:
        output_file.write(
            to_bytes(
                unicode_content,
                to_encoding=to_encoding))
    #


# vim:fileencoding=utf-8 autoindent ts=4 sw=4 sts=4 expandtab:



# -*- coding: utf-8 -*-

"""

smallparts.namespaces - simple dict-based namespaces

------------------------------------------------------------------------

Adapted from:

ActiveState Code » Recipes » A Simple Namespace Class (Python recipe)
<http://code.activestate.com/recipes/577887-a-simple-namespace-class/>

------------------------------------------------------------------------

"""


#
# Classes
#


class Namespace(dict):

    """A dict subclass that exposes its items as attributes.

    Warning: Namespace instances only have direct access to the
    attributes defined in the visible_attributes tuple
    """

    visible_attributes = ('items', )

    def __repr__(self):
        """Object representation"""
        return '{0}({1})'.format(
            type(self).__name__,
            super(Namespace, self).__repr__())

    def __dir__(self):
        """Members sequence"""
        return tuple(self)

    def __getattribute__(self, name):
        """Access a visible attribute
        or return an existing dict member
        """
        if name in type(self).visible_attributes:
            return object.__getattribute__(self, name)
        #
        try:
            return self[name]
        except KeyError:
            raise AttributeError(
                '{0!r} object has no attribute {1!r}'.format(
                    type(self).__name__, name))
        #

    def __setattr__(self, name, value):
        """Set an attribute"""
        self[name] = value

    def __delattr__(self, name):
        """Delete an attribute"""
        del self[name]


class DefaultNamespace(Namespace):

    """Namespace object with a default value for non-existent
    attributes
    """

    visible_attributes = ('items', 'default__value__')

    def __init__(self, *args, **kwargs):
        """Initialize like a dict, but steal the 'default'
        keyword argument if given"""
        object.__setattr__(self,
                           'default__value__',
                           kwargs.pop('default', None))
        super(DefaultNamespace, self).__init__(*args, **kwargs)

    def __getattribute__(self, name):
        """Access a visible attribute
        or return the default value (and cache it)
        """
        if name in type(self).visible_attributes:
            return object.__getattribute__(self, name)
        #
        return dict.setdefault(self, name, self.default__value__)


class EnhancedNamespace(Namespace):

    """Namespace object with additional constructor methods
    for using only selected names
    """

    visible_attributes = ('values', 'items')

    @classmethod
    def from_object(cls, object_, names=None):
        """Try to construct a Namespace object from an arbitrary object"""
        if names is None:
            names = dir(object_)
        #
        return cls((name, getattr(object_, name)) for name in names)

    @classmethod
    def from_mapping(cls, mapping, names=None):
        """Construct a Namespace object from a mapping (i.e a dict)"""
        if names:
            return cls((name, mapping[name]) for name in names)
        #
        return cls(mapping)

    @classmethod
    def from_sequence(cls, sequence, names=None):
        """Construct a Namespace object from a sequence of 2-tuples"""
        if names:
            return cls((name, value) for (name, value) in sequence
                       if name in names)
        #
        return cls(sequence)


class InstantNames(Namespace):

    """Namespace object where instance attributes are defined
    on access, derived from the given name with the translation
    functions applied on it
    """

    translations = 'translation_functions'

    def __init__(self, *translation_functions, **set_values_directly):
        """Register translation functions
        and set the given values directly
        """
        object.__setattr__(self,
                           type(self).translations,
                           translation_functions)
        super(InstantNames, self).__init__(set_values_directly)

    def __repr__(self):
        """Object representation"""
        instance_attributes_list = [
            repr(function) for function
            in object.__getattribute__(self, type(self).translations)]
        instance_attributes_list.extend(
            '{0}={1!r}'.format(name, self[name]) for name in self)
        return '{0}({1})'.format(
            type(self).__name__,
            ', '.join(instance_attributes_list))

    def __getattribute__(self, name):
        """Access an allowed internal attribute
        or set the instance attribute
        """
        if name in (type(self).translations,):
            return object.__getattribute__(self, name)
        #
        try:
            attribute_value = self[name]
        except KeyError:
            attribute_value = name
            for translation in object.__getattribute__(
                    self, type(self).translations):
                attribute_value = translation(attribute_value)
            self[name] = attribute_value
        return attribute_value


# vim:fileencoding=utf-8 autoindent ts=4 sw=4 sts=4 expandtab:



# -*- coding: utf-8 -*-

"""

smallparts.text.join

Text joining functions where the parts are given as positional arguments

"""


def by_blanks(*parts):
    """return parts joined by blanks"""
    return ' '.join(parts)


def directly(*parts):
    """return parts joined directly"""
    return ''.join(parts)


def by_newlines(*parts):
    """return parts joined by newline characters - Unix style"""
    return '\n'.join(parts)


def by_crlf(*parts):
    """return parts joined by crlf - DOS style"""
    return '\r\n'.join(parts)


def using(join_string, *parts):
    """return parts joined using the given string"""
    return join_string.join(parts)


# vim:fileencoding=utf-8 autoindent ts=4 sw=4 sts=4 expandtab:



# -*- coding: utf-8 -*-

"""

smallparts.l10n.languages

(Natural) language definitions

"""


DE = 'de'
EN = 'en'
ES = 'es'
FR = 'fr'

DEFAULT = EN
SUPPORTED = {DE, EN, ES, FR}


def missing_translation(lang, message=None):
    """Return an explanation if a language is not supported at all
    or if the translation is missing, to be used as a message for
    a ValuError which is raised in that case.
    """
    if lang not in SUPPORTED:
        return 'Language {0!r} not supported!'.format(lang)
    #
    return message or 'No {0!r} translation available yet!'.format(lang)


# vim:fileencoding=utf-8 autoindent ts=4 sw=4 sts=4 expandtab:



# -*- coding: utf-8 -*-

"""

smallparts.sequences

Utility functions for sequences

"""


DEFAULT_JOINER = ','
EMPTY = ''


def flatten(iterable, depth=None):
    """Flatten the given iterable recursively and return a list."""
    if isinstance(iterable, (str, bytes)):
        return [iterable]
    #
    if depth is None:
        children_depth = None
    else:
        children_depth = depth - 1
        if depth < 0:
            return [iterable]
        #
    #
    flattened_list = []
    try:
        for item in iterable:
            flattened_list.extend(flatten(item, depth=children_depth))
        #
    except TypeError:
        return [iterable]
    else:
        return flattened_list
    #


def raw_join(iterable,
             prefix=None,
             joiner=DEFAULT_JOINER,
             final_joiner=None,
             suffix=None):
    """Return a unicode string containing the list items
    joined together according to the provided parameters
    """
    final_joiner = final_joiner or joiner
    words_sequence = [str(item) for item in iterable]
    items_list = words_sequence[:-2]
    items_list.append(final_joiner. join(words_sequence[-2:]))
    return EMPTY.join((prefix or EMPTY,
                       joiner.join(items_list),
                       suffix or EMPTY))


# vim:fileencoding=utf-8 autoindent ts=4 sw=4 sts=4 expandtab:



# -*- coding: utf-8 -*-

"""

smallparts.markup.parsers

Markup parsing

"""


import html.parser
import re

from smallparts import constants


#
# Constants
#

# Keyword constants for the HtmlTagStripper class

IMAGES_WITH_ALT_TEXT_ONLY = 'images with alt text only'


#
# Classes
#


class EntityResolver():

    """Resolve entities"""

    # Map xml emtity names to codepoints
    xml_name2codepoint = {
        'lt': 60,
        'gt': 62,
        'amp': 38,
        'apos': 39,
        'quot': 34
    }
    # Numeric character references:
    # <https://www.w3.org/TR/xml/#NT-CharRef>
    re_charref = r'#(?:[0-9]+|x[0-9a-fA-F]+)'
    # Entity names: see <https://www.w3.org/TR/xml/#NT-NameStartChar>
    # and https://www.w3.org/TR/xml/#NT-NameChar
    re_name_start_char = (
        r':A-Z_a-z\xc0-\xd6\xd8-\xf6\xf8-\u02ff\u0370-\u037d\u037f-\u1fff'
        r'\u200c-\u200d\u2070-\u218f\u2c00-\u2fef\u3001-\ud7ff'
        r'\uf900-\ufdcf\ufdf0-\ufffd\U00010000-\U000effff')
    re_name_char = '-' + re_name_start_char + \
        r'\.0-9\xb7\u0300-\u036f\u203f-\u2040'

    def __init__(self, named_entities=None):
        """Precompile the regular expression matching all entity references.
        Allocate the internal named entities mapping from the provided
        named_entities mapping, default to XML_NAME2CODEPOINT.
        """
        self.__prx_entity = re.compile(
            '&([{0}][{1}]+|{2});'.format(
                self.re_name_start_char,
                self.re_name_char,
                self.re_charref))
        self.__named_entities = {}
        try:
            self.set_named_entities(named_entities)
        except TypeError:
            self.set_named_entities(self.xml_name2codepoint)
        #

    def set_named_entities(self, named_entities):
        """Override the internal named entities mapping with
        the provided dict after checking its contents for validity.
        Raises a TypeError if anything other than a mapping was
        provided.
        Raises a ValueError if any invalid replacement is
        defined in the provided mapping.
        """
        try:
            names_and_replacements = named_entities.items()
        except AttributeError:
            raise TypeError('Please provide a mapping.')
        #
        checked_replacements = {}
        for (name, replacement) in names_and_replacements:
            if isinstance(replacement, str):
                checked_replacements[name] = replacement
            elif isinstance(replacement, int):
                # Might raise a ValueError on invalid values
                # outside the range 0–1114111 (0x0–0x10ffff)
                checked_replacements[name] = chr(replacement)
            else:
                raise ValueError(
                    'Only integers up to 0x10ffff or strings'
                    ' are allowed as replacements.')
            #
        #
        self.__named_entities = checked_replacements

    @staticmethod
    def resolve_charref(charref):
        """Resolve a numeric (hexadecimal or decimal) character reference"""
        if charref.startswith('x'):
            codepoint = int('0' + charref, 16)
        else:
            codepoint = int(charref)
        return chr(codepoint)

    def resolve_named_entity(self, name):
        """Resolve a named entity from the internal mapping"""
        try:
            return self.__named_entities[name]
        except KeyError:
            raise ValueError('Name {0!r} not defined.'.format(name))
        #

    def resolve_any_entity(self, entityref):
        """Dispatch to the appropriate resolver method"""
        if entityref.startswith('#'):
            return self.resolve_charref(entityref[1:])
        #
        return self.resolve_named_entity(entityref)

    def __resolve_match(self, match_object):
        """Return the resolved matched entity's replacement,
        or the entity itself if it could not be resolved
        """
        try:
            return self.resolve_any_entity(match_object[1])
        except ValueError:
            return match_object[0]
        #

    def resolve_all_entities(self, source_text):
        """Resolve all entities in source_text using
        this object's __resolve_match() method as replacement
        """
        return self.__prx_entity.sub(self.__resolve_match,
                                     source_text)


# pylint: disable=abstract-method ; The error method is not used


class HtmlTagStripper(html.parser.HTMLParser):

    """Return only the data, concatenated using constants.EMPTY,
    with whitespace squeezed together, but preserving line breaks.
    """

    image_placeholder_with_alt_text = '[image: {0[alt]}]'
    image_placeholder_empty = '[image]'

    re_multiple_space = r'[ \t\r\f\v]{2,}'
    re_newline_and_whitespace = r'\s*\n\s*'

    treated_as_block = {
        'address',
        'article',
        'aside',
        'bdi',
        'blockquote',
        'body',
        'br',
        'canvas',
        'caption',
        'datalist',
        'details',
        'dialog',
        'dir',
        'div',
        'dl',
        'dt',
        'fieldset',
        'figure',
        'figcaption',
        'footer',
        'form',
        'h1',
        'h2',
        'h3',
        'h4',
        'h5',
        'h6',
        'head',
        'header',
        'hr',
        'html',
        'iframe',
        'legend',
        'li',
        'main',
        'menu',
        'nav',
        'noscript',
        'ol',
        'option',
        'output',
        'p',
        'picture',
        'pre',
        'ruby',
        'rt',
        'rb',
        'script',
        'section',
        'select',
        'slot',
        'style',
        'summary',
        'table',
        'tbody',
        'template',
        'tfoot',
        'thead',
        'title',
        'tr',
        'ul',
    }

    def __init__(self, image_placeholders=IMAGES_WITH_ALT_TEXT_ONLY):
        """Instantiate the base class and define instance variables"""
        html.parser.HTMLParser.__init__(self, convert_charrefs=True)
        if image_placeholders:
            if image_placeholders == IMAGES_WITH_ALT_TEXT_ONLY:
                self.image_placeholder_empty = ''
            #
        else:
            self.image_placeholder_with_alt_text = ''
        #
        self.__content_list = []
        self.__images = []
        self.__in_body = False
        self.__prx_multiple_space = re.compile(self.re_multiple_space)
        self.__prx_newline_and_whitespace = re.compile(
            self.re_newline_and_whitespace, re.DOTALL)

    def __add_body_content(self, content):
        """Add content if self.__in_body"""
        if self.__in_body:
            self.__content_list.append(content)
        #

    def __add_whitespace(self, lowercased_tag):
        """Add whitespace
        (newline if the tag is treated as a block element,
         else a blank)
        """
        if lowercased_tag in self.treated_as_block:
            self.__add_body_content(constants.NEWLINE)
        else:
            self.__add_body_content(constants.BLANK)
        #

    @property
    def content(self):
        """Return the result"""
        collected_content = self.__prx_multiple_space.sub(
            constants.BLANK,
            constants.EMPTY.join(self.__content_list))
        return self.__prx_newline_and_whitespace.sub(
            constants.NEWLINE,
            collected_content).strip()

    @property
    def images(self):
        """Return the saved image data"""
        return list(self.__images)

    def handle_data(self, data):
        """Collect content"""
        self.__add_body_content(data)

    def handle_starttag(self, tag, attrs):
        """Handle a start tag"""
        tag = tag.lower()
        self.__add_whitespace(tag)
        if tag == 'body':
            self.__in_body = True
        elif tag == 'img':
            # save images' attributes
            current_image = dict(attrs)
            self.__images.append(current_image)
            try:
                self.__add_body_content(
                    self.image_placeholder_with_alt_text.format(
                        current_image))
            except KeyError:
                self.__add_body_content(self.image_placeholder_empty)
            #
        #

    def handle_endtag(self, tag):
        """Handle an end tag"""
        tag = tag.lower()
        self.__add_whitespace(tag)
        if tag == 'body':
            self.__in_body = False
        #


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-

"""

smallparts.text.reduce

Classes, functions and rules for reducing unicode text to ASCII

"""


from smallparts import constants


#
# Reduction rules as dicts: {source_characters: ascii_replacement, …}
#

LATIN = {
    # Latin characters from the
    # Latin-1 supplement (U0080–U00ff) and
    # Latin extended-A (U0100–U017f) Unicode blocks
    #
    'ÀÁÂÃÄÅĀĂĄ': 'A',
    'Æ': 'Ae',
    'ÇĆĈĊČ': 'C',
    'Ď': 'D',
    'ÐĐ': 'Dh',
    'ÈÉÊËĒĔĖĘĚ': 'E',
    'ĜĞĠĢ': 'G',
    'ĤĦ': 'H',
    'ÌÍÎÏĨĪĬĮİ': 'I',
    'Ĳ': 'IJ',
    'Ĵ': 'J',
    'Ķ': 'K',
    'ĹĻĽĿŁ': 'L',
    'ÑŃŅŇ': 'N',
    'Ŋ': 'Ng',
    'ÒÓÔÕÖŌŎŐØ': 'O',
    'Œ': 'Oe',
    'ŔŖŘ': 'R',
    'ŚŜŞŠ': 'S',
    'ŢŤŦ': 'T',
    'Þ': 'Th',
    'ÙÚÛÜŨŪŬŮŰŲ': 'U',
    'Ŵ': 'W',
    'ÝŶŸ': 'Y',
    'ŹŻŽ': 'Z',
    'àáâãäåāăą': 'a',
    'æ': 'ae',
    'çćĉċč': 'c',
    'ď': 'd',
    'ðđ': 'dh',
    'èéêëēĕėęě': 'e',
    'ĝğġģ': 'g',
    'ĥħ': 'h',
    'ìíîïĩīĭįı': 'i',
    'ĳ': 'ij',
    'ĵ': 'j',
    'ķĸ': 'k',
    'ĺļľŀł': 'l',
    'ñńņň': 'n',
    'ŋ': 'ng',
    'òóôõöōŏőø': 'o',
    'œ': 'oe',
    'ŕŗř': 'r',
    'śŝşšſ': 's',
    'ß': 'ss',
    'ţťŧ': 't',
    'þ': 'th',
    'ùúûüũūŭůűų': 'u',
    'ŵ': 'w',
    'ýŷÿ': 'y',
    'źżž': 'z'
}

PUNCTUATION = {
    # Punctuation and symbols from the
    # Latin-1 supplement (U0080–U00ff) and
    # General punctuation (U2000–U206f) Unicode blocks
    #
    # Spacing characters → space
    '\u0080\u2000\u2001\u2002\u2003\u2004'
    '\u2005\u2006\u2007\u2008\u2009\u200a': '\x20',
    # Soft hyphen → hyphen in parentheses
    '\u00ad': '(-)',
    # Hyphen bullet → hyphen
    '\u2043': '-',
    # Dashes → single, double or triple hyphen
    '\u2010\u2011': '-',
    '\u2012\u2013': '--',
    '\u2014\u2015': '---',
    # Double vertical line → double pipe
    '\u2016': '||',
    # Double low line → underscore
    '\u2017': '_',
    # Quotation marks → apostrophe, quotation mark, << or >>
    '\u2018\u2019\u201a\u201b': '\x27',
    '\u201c\u201d\u201e\u201f': '"',
    '«': '<<',
    '»': '>>',
    # Daggers → plus sign(s)
    '\u2020': '+',
    '\u2021': '++',
    # Leader dots, ellipsis → dots
    '\u2024': '.',
    '\u2025': '..',
    '\u2026': '...',
    # Bullets, middle dots, times sign → asterisk
    '\u00b7\u00d7\u2022\u2027\u204c\u204d\u204e': '*',
    # Line and paragraph separators → ASCII LF
    '\u2028': '\n',
    '\u2029': '\n\n',
    # Per mille and per myriad (= per then thousand) signs → {description}
    '\u2030': '{permille}',
    '\u2031': '{permyriad}',
    # Primes → apostrophes, reverse primes → grave accents
    '\u2032': '\x27',
    '\u2033': '\x27\x27',
    '\u2034': '\x27\x27\x27',
    '\u2057': '\x27\x27\x27\x27',
    '\u2035': '\x60',
    '\u2036': '\x60\x60',
    '\u2037': '\x60\x60\x60',
    # Caret, angle quotation marks → circumflex, less-than, greater than
    '\u2038': '^',
    '\u2039': '<',
    '\u203a': '>',
    # Exclamation and question marks, semicolon
    '¡': '!',
    '¿': '?',
    '\u203c': '!!',
    '\u203d': '?!',
    '\u2047': '??',
    '\u2048': '?!',
    '\u2049': '!?',
    '\u204f': ';',
    # Division sign and fraction slash → slash
    '÷\u2044': '/',
    # Tironian sign et → ampersand
    '\u204a': '&',
    # Various punctuation and symbols from the U2000 block
    '\u204b': '{reversed pilcrow}\n',
    '\u2051': '**',
    '\u2052': './.',
    '\u2053': '~',
    '\u2055': '*',
    # Various punctuation and symbols from the U0080 block
    '¢': 'ct',
    '£': 'GBP',
    '¤': '{currency}',
    '¥': 'JPY',
    '¦': '|',
    '§': '{section sign}',
    '¨': '"',
    '©': '(C)',
    'ª': '^a',
    '¬': '{not}',
    '®': '(R)',
    '¯': '{macron}',
    '°': '{degree}',
    '±': '+-',
    '¹': '^1',
    '²': '^2',
    '³': '^3',
    '´': '\x27',
    'µ': '{micro}',
    '¶': '{pilcrow}\n',
    '¸': '{cedilla}',
    'º': '^o',
    '¼': '1/4',
    '½': '1/2',
    '¾': '3/4'
}

ISO_CURRENCY = {
    # ISO 4217 codes for all currency symbols from the
    # Currency symbols (U20a0–U20bf) Unicode block
    # that are clearly attributable
    #
    '₠': 'ECU',
    '₣': 'FRF',
    '₦': 'NGN',
    '₧': 'ESP',
    '₪': 'ILS',
    '₫': 'VND',
    '€': 'EUR',
    '₭': 'LAK',
    '₮': 'MNT',
    '₯': 'GRD',
    '₱': 'PHP',
    '₲': 'PYG',
    '₳': 'ARA',
    '₴': 'UAH',
    '₵': 'GHS',
    '₸': 'KZT',
    '₹': 'INR',
    '₺': 'TRY',
    '₼': 'AZN',
    '₽': 'RUB',
    '₾': 'GEL',
    '₿': 'BTC',
    'ƒ': 'NLG',
    '฿': 'THB',
    '৳': 'BDT'
}

NON_ISO_CURRENCY = {
    # Names for all currency symbols from the
    # Currency symbols (U20a0–U20bf) Unicode block that are
    # NOT clearly attributable or do not have a ISO 4217 code
    #
    '₡': '{Colon}',             # CRC and SVC
    '₢': '{Cruzeiro}',          # BRB, BRC, BRN, BRE, BRR
    '₤': '{Lira}',              # ITL, MTL, SML, VAL, possybly also SYP
    '₥': '{Mill}',              # former US currency unit (1/1000 $)
    '₨': '{Rupee}',             # various currencies; Indian Rupee: see INR
    '₩': '{Won}',               # KPW and KRW
    '₰': '{Pfennig}',           # former German curreny unit (1/100 Mark)
    '₶': '{Livre tournois}',    # former French currency, 13th to 18th century
    '₷': '{Spesmilo}',          # historical proposed int'l currency
    '₻': '{Nordic Mark}'        # Danish rigsdaler
}

GERMAN_OVERRIDES = {
    # German-language overrides including U1e9e (ẞ, capital ß)
    #
    'Ä': 'Ae',
    'ÖŐØ': 'Oe',
    'ẞ': 'SZ',
    'ÜŰ': 'Ue',
    'ä': 'ae',
    'öőø': 'oe',
    'üű': 'ue',
    '\u2030': '{Promille}',
    '¤': '{Waehrung}',
    '§': 'Par.',
    '¬': '{nicht}',
    '°': '{Grad}',
    '¶': '{Absatzmarke}\n',
    '¸': '{Cedille}',
    '₰': 'Pf.'
}


#
# Helper functions
#


def checked_ascii(unicode_text):
    """If the unicode_text is ASCII only (encodable unsing 'ascii'),
    return it unchanged.
    Raise a ValueError om all encoding errors,
    or a TypeError if no string was provided.
    """
    try:
        unicode_text.encode(constants.ASCII)
    except AttributeError:
        raise TypeError('Expected a str object')
    except UnicodeEncodeError:
        raise ValueError(
            '{0} is not ASCII only'.format(unicode_text))
    #
    return unicode_text


#
# Classes
#


class ConversionTable:

    """Conversion table for reductions to ASCII"""

    max_c1_control = '\x9f'

    def __init__(self,
                 rules_mapping,
                 default_replacement=None,
                 remove_c1_controls=False):
        """Set up the internal conversion mapping
        from the given rules mapping
        (key: string of characters to convert; value: ASCII replacement)
        """
        self.__reductions = {}
        self.add_reduction_items(rules_mapping.items())
        try:
            self.__default_replacement = checked_ascii(default_replacement)
        except TypeError:
            self.__default_replacement = None
        #
        self.remove_c1_controls = remove_c1_controls

    @property
    def reduction_items(self):
        """self.__reductions.items() (readonly access)"""
        return self.__reductions.items()

    @property
    def default_replacement(self):
        """Read only access to self.__default_replacement"""
        return self.__default_replacement

    def add_reduction_items(self, reduction_items):
        """Add the given items (= souce_characters, replacement tuples)
        to the internal mapping
        """
        for source_characters, replacement in reduction_items:
            valid_replacement = checked_ascii(replacement)
            for character in source_characters:
                self.__reductions[character] = valid_replacement
            #
        #

    def __add__(self, other):
        """Return a new object as a copy, with the internal mapping
        updated form the added object's mapping
        """
        if isinstance(other, ConversionTable):
            items_to_add = other.reduction_items
        elif isinstance(other, dict):
            items_to_add = other.items()
        else:
            raise TypeError('Cannot add any other type than'
                            ' ConversionTable or dict!')
        #
        result = ConversionTable(
            self.__reductions,
            default_replacement=self.__default_replacement,
            remove_c1_controls=self.remove_c1_controls)
        result.add_reduction_items(items_to_add)
        return result

    __radd__ = __add__

    def __eq__(self, other):
        """Return True if reductions and default replacements
        of both ConversionTables are equal
        """
        return (self.reduction_items == other.reduction_items and
                self.default_replacement == other.default_replacement)

    def reduce_character(self, character):
        r"""Reduce a single unicode character to ASCII
        according to the following rules:
        - if the character is valid ASCII, return it unchanged
        - if it is in the C1 controls range (U0080-U009f)
          AND the ConversionTable instance was created
          with remove_c1_controls=True, return an empty string
        - in any other case, return the defined replacement.
        - if no replacement is defined, either return the
          default replacement or, if no default replacement
          had been specified, the appropriate Unicode escape
          (\xNN, \uNNNN or \uNNNNNNNN).
        """
        try:
            return checked_ascii(character)
        except ValueError:
            if character <= self.max_c1_control and self.remove_c1_controls:
                return constants.EMPTY
            #
        #
        try:
            return self.__reductions[character]
        except KeyError:
            if self.__default_replacement is None:
                return character.encode(
                    'unicode_escape').decode(constants.ASCII)
            #
            return self.__default_replacement
        #

    def reduce_text(self, unicode_text):
        """Reduce the provided unicode text character by character"""
        return constants.EMPTY.join(self.reduce_character(character)
                                    for character in unicode_text)


#
# End of classes, start of functions
#


def latin_to_ascii(unicode_text, *additional_rules):
    """Reduce the given text to ASCII using basic latin rules
    plus the additional rules given as positional parameters
    after the text
    """
    conversion_table = ConversionTable(LATIN)
    for rule in additional_rules:
        conversion_table = conversion_table + rule
    #
    return conversion_table.reduce_text(unicode_text)


# vim:fileencoding=utf-8 autoindent ts=4 sw=4 sts=4 expandtab:



# -*- coding: utf-8 -*-

"""

smallparts.text.split

Text splitting functions

"""


def lines_for_reconstruction(unicode_text):
    """Split unicode_text using the splitlines() str method,
    but append an empty string at the end if the last line
    of the original text ends with a line break, in order
    to be able to keep this trailing line end when applying
    LINE_BREAK.join(splitted_lines).
    The line break characters below were taken from
    <https://docs.python.org/3/library/stdtypes.html#str.splitlines>
    """
    if isinstance(unicode_text, str):
        splitted_lines = unicode_text.splitlines()
    else:
        raise TypeError('This function requires a unicode argument.')
    #
    if unicode_text and \
            unicode_text[-1] in '\n\r\v\f\x1c\x1d\x1e\x85\u2028\u2029':
        splitted_lines.append('')
    #
    return splitted_lines


# vim:fileencoding=utf-8 autoindent ts=4 sw=4 sts=4 expandtab:



# -*- coding: utf-8 -*-

"""

smallparts.constants - common constants

"""


#
# Single character constants
#

AMPERSAND = '&'
ASTERISK = '*'
AT = '@'
BLANK = SPACE = SP = ' '
BRACE_OPEN = '{'
BRACE_CLOSE = '}'
COLON = ':'
COMMA = ','
CARRIAGE_RETURN = CR = '\r'
DASH = '-'
DOT = '.'
DOUBLE_QUOTE = '"'
EMPTY = ''
EQUALS = '='
HASH = POUND = '#'
LINEFEED = LF = NEWLINE = NL = '\n'
PERCENT = '%'
PIPE = '|'
PLUS_SIGN = '+'
QUESTION_MARK = '?'
SEMICOLON = ';'
SINGLE_QUOTE = "'"
SLASH = '/'
TILDE = '~'
UNDERSCORE = '_'

#
# Compound constants
#

COLON_BLANK = COLON + BLANK
COMMA_BLANK = COMMA + BLANK
CRLF = CR + LF

#
# Numeric constants
#

ZERO = 0
ONE = 1
FIRST_INDEX = ZERO
SECOND_INDEX = ONE
LAST_INDEX = -1

#
# Encodings
#

ASCII = 'ascii'
CP1252 = 'cp1252'
UTF_8 = 'utf-8'

#
# Functional constants
#

MODE_APPEND = 'a+'
MODE_APPEND_BINARY = 'a+b'
MODE_READ = 'r'
MODE_READ_BINARY = 'rb'
MODE_WRITE = 'w'
MODE_WRITE_BINARY = 'wb'

YES = 'yes'
NO = 'no'

XML_1_0 = '1.0'
XML_1_1 = '1.1'

#
# Return codes
#

RC_ERROR = 1
RC_OK = 0


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-

"""

smallparts.text.templates

Enhanced String Template class

"""

import string


class EnhancedStringTemplate(string.Template):

    """string.Template subclass adding one property:
    the set of variable names from the template
    """

    @property
    def variable_names(self):
        """Return the set of variable names in the template"""
        result = set()
        # pylint: disable=no-member ; false positive on Template.pattern
        for placeholder_match in self.pattern.finditer(self.template):
            placeholder_name = placeholder_match.group('named') or \
                placeholder_match.group('braced')
            if placeholder_name:
                result.add(placeholder_name)
            #
        #
        return result


# vim:fileencoding=utf-8 autoindent ts=4 sw=4 sts=4 expandtab:



# -*- coding: utf-8 -*-

"""

smallparts.text.translate - text translation functions

"""

from smallparts import constants


def remove_trailing_underscores(name):
    """Remove trailing underscores"""
    return name.rstrip(constants.UNDERSCORE)


def underscores_to_blanks(name):
    """translate underscores to blanks"""
    return name.replace(constants.UNDERSCORE, constants.BLANK)


def underscores_to_dashes(name):
    """translate underscores to dashes"""
    return name.replace(constants.UNDERSCORE, constants.DASH)


# vim:fileencoding=utf-8 autoindent ts=4 sw=4 sts=4 expandtab:



# -*- coding: utf-8 -*-

"""

smallparts.time_display - time representation functions

"""


from smallparts.l10n import enumerations as ln_enum
from smallparts.l10n import time_indications as ln_time


FS_DATE = '%Y-%m-%d'
FS_TIME = '%H:%M:%S'
FS_DATETIME = '{0} {1}'.format(FS_DATE, FS_TIME)
FS_USEC = '{0}.%f'
FS_MSEC = '{0}.{1:03d}'
FS_DATETIME_WITH_USEC = FS_USEC.format(FS_DATETIME)
FS_TIME_WITH_USEC = FS_USEC.format(FS_TIME)

SECONDS_PER_MINUTE = 60
MINUTES_PER_HOUR = 60
HOURS_PER_DAY = 24
DAYS_PER_WEEK = 7

DEFAULT_DISPLAY_LIMITS = {
    ln_time.SECONDS: 60 * SECONDS_PER_MINUTE,
    ln_time.MINUTES: 24 * MINUTES_PER_HOUR,
    ln_time.HOURS: 7 * HOURS_PER_DAY,
    ln_time.DAYS: 4 * DAYS_PER_WEEK}


def _as_specified(datetime_object,
                  format_string=FS_DATETIME,
                  with_msec=False,
                  with_usec=False):
    """Return the datetime object formatted as specified"""
    if with_usec:
        return datetime_object.strftime(FS_USEC.format(format_string))
    #
    if with_msec:
        msec = datetime_object.microsecond // 1000
        return FS_MSEC.format(datetime_object.strftime(format_string), msec)
    # implicit else:
    return datetime_object.strftime(format_string)


def as_date(datetime_object):
    """Return the datetime object formatted as date"""
    return datetime_object.strftime(FS_DATE)


def as_datetime(datetime_object, with_msec=False, with_usec=False):
    """Return the datetime object formatted as datetime,
    convenience wrapper around _as_specified
    """
    return _as_specified(datetime_object,
                         format_string=FS_DATETIME,
                         with_msec=with_msec,
                         with_usec=with_usec)


def as_time(datetime_object, with_msec=False, with_usec=False):
    """Return the datetime object formatted as time,
    convenience wrapper around _as_specified
    """
    return _as_specified(datetime_object,
                         format_string=FS_TIME,
                         with_msec=with_msec,
                         with_usec=with_usec)


def pretty_printed_timedelta(timedelta, limits=None, lang=None):
    """Return the timedelta, pretty printed
    and regarding the limits per time component
    """
    result = []
    limits = limits or {}
    totals = {}
    values = {}
    totals[ln_time.SECONDS] = int(timedelta.total_seconds())
    previous_unit = ln_time.SECONDS
    for current_unit, conversion_factor in (
            (ln_time.MINUTES, SECONDS_PER_MINUTE),
            (ln_time.HOURS, MINUTES_PER_HOUR),
            (ln_time.DAYS, HOURS_PER_DAY),
            (ln_time.WEEKS, DAYS_PER_WEEK)):
        totals[current_unit], values[previous_unit] = divmod(
            totals[previous_unit], conversion_factor)
        if values[previous_unit] and \
                totals[previous_unit] < limits.get(
                        previous_unit,
                        DEFAULT_DISPLAY_LIMITS[previous_unit]):
            tc_kwargs = {previous_unit: values[previous_unit],
                         'lang': lang}
            result.append(
                ln_time.pretty_print_component(**tc_kwargs))
        #
        if not totals[current_unit]:
            break
        previous_unit = current_unit
    #
    if totals.get(ln_time.WEEKS):
        result.append(
            ln_time.pretty_print_component(weeks=totals[ln_time.WEEKS],
                                           lang=lang))
    #
    return ln_enum.enumeration(result[::-1], ln_enum.AND, lang=lang)


# vim:fileencoding=utf-8 autoindent ts=4 sw=4 sts=4 expandtab:



# -*- coding: utf-8 -*-

"""

smallparts.l10n.enumerations

Natural language enumerations

"""


from smallparts.l10n import languages

from smallparts.sequences import raw_join


BLANK = ' '
COMMA = ','
EMPTY = ''

AND = 'and'
OR = 'or'
EITHER = 'either'
NEITHER = 'neither'

DEFAULT = 'default'
EXCEPTIONS = 'exceptions'

BEFORE = 'before'
AFTER = 'after'

SUPPORTED_ENUMS = (AND, OR, EITHER, NEITHER)

# (prefix, joiner, last joiner)
ENUM_JOINERS = {
    languages.EN: {
        AND: (None, COMMA, AND),
        OR: (None, COMMA, OR),
        EITHER: (EITHER, COMMA, OR),
        NEITHER: (NEITHER, COMMA, 'nor')
    },
    languages.DE: {
        AND: (None, COMMA, 'und'),
        OR: (None, COMMA, 'oder'),
        EITHER: ('entweder', COMMA, 'oder'),
        NEITHER: ('weder', COMMA, 'noch')
    },
    languages.ES: {
        AND: (None, COMMA, 'y'),
        OR: (None, COMMA, 'o'),
        EITHER: ('ya sea', COMMA, 'o'),
        NEITHER: ('ni', COMMA, 'ni')
    },
    languages.FR: {
        AND: (None, COMMA, 'et'),
        OR: (None, COMMA, 'ou'),
        EITHER: ('soit', COMMA, 'ou'),
        NEITHER: ('ni', COMMA, 'ni')
    },
    '__test__': {AND: (None, None, None)}
}

# Spacing rules: (before, after)
SPACING_RULES = {
    languages.EN: {
        DEFAULT: (True, True),
        EXCEPTIONS: {BEFORE: {',;.:!?/': False},
                     AFTER: {'/': False}}
    },
    languages.DE: {
        DEFAULT: (True, True),
        EXCEPTIONS: {BEFORE: {',;.:!?': False}}
    },
    languages.ES: {
        DEFAULT: (True, True),
        EXCEPTIONS: {BEFORE: {',;.:!?': False}}
    },
    languages.FR: {
        DEFAULT: (True, True),
        EXCEPTIONS: {}
    },
    '__test__': {
        DEFAULT: (False, False),
        EXCEPTIONS: {}
    }
}

#
# Functions
#


def apply_spacing_rules(joiner, lang=None):
    """Apply language-specific spacing rules"""
    stripped_joiner = joiner.strip()
    if not stripped_joiner:
        return joiner
    #
    lang = lang or languages.DEFAULT
    try:
        spacing_rules = SPACING_RULES[lang]
    except KeyError:
        raise ValueError(
            languages.missing_translation(
                lang,
                message='No spacing rules available yet'
                'for {0!r}!'.format(lang)))
    #
    for (characters, rule_before) in spacing_rules[EXCEPTIONS].get(
            BEFORE, {}).items():
        if stripped_joiner[0] in characters:
            space_before = rule_before
            break
        #
    else:
        space_before = spacing_rules[DEFAULT][0]
    #
    for (characters, rule_after) in spacing_rules[EXCEPTIONS].get(
            AFTER, {}).items():
        if stripped_joiner[-1] in characters:
            space_after = rule_after
            break
        #
    else:
        space_after = spacing_rules[DEFAULT][-1]
    #
    prefix = suffix = EMPTY
    if space_before:
        prefix = BLANK
    #
    if space_after:
        suffix = BLANK
    #
    return EMPTY.join((prefix, stripped_joiner, suffix))


def enumeration(sequence, enum_type, lang=None):
    """Return the sequence enumerated according to the
    given enum_type and language
    """
    lang = lang or languages.DEFAULT
    try:
        enum_joiners = ENUM_JOINERS[lang]
    except KeyError:
        raise ValueError(languages.missing_translation(lang))
    else:
        try:
            prefix, joiner, final_joiner = enum_joiners[enum_type]
        except KeyError:
            raise ValueError(
                'No {0!r} translation available for {1!r}!'.format(
                    lang, enum_type))
        #
    #
    if joiner is None:
        joiner = COMMA
    #
    if final_joiner is None:
        final_joiner = joiner
    #
    if prefix:
        prefix = prefix.strip() + BLANK
    #
    return raw_join(
        sequence,
        prefix=prefix,
        joiner=apply_spacing_rules(joiner, lang=lang),
        final_joiner=apply_spacing_rules(final_joiner, lang=lang))


# vim:fileencoding=utf-8 autoindent ts=4 sw=4 sts=4 expandtab:



# -*- coding: utf-8 -*-

"""

smallparts.markup.elements

Markup (HTML, XML) generation:
Simple XML, XHTML 1.0 and HTML 5 element definitions

"""


import xml.sax.saxutils

from smallparts import constants

from smallparts.text import translate


#
# Constants
#

# From the XHTML Strict DTD
# <https://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd>
XHTML_1_0_STRICT = {
    'a',
    'abbr',
    'acronym',
    'address',
    'area',
    'b',
    'base',
    'bdo',
    'big',
    'blockquote',
    'body',
    'br',
    'button',
    'caption',
    'cite',
    'code',
    'col',
    'colgroup',
    'dd',
    'del',
    'dfn',
    'div',
    'dl',
    'dt',
    'em',
    'fieldset',
    'form',
    'h1',
    'h2',
    'h3',
    'h4',
    'h5',
    'h6',
    'head',
    'hr',
    'html',
    'i',
    'img',
    'input',
    'ins',
    'kbd',
    'label',
    'legend',
    'li',
    'link',
    'map',
    'meta',
    'noscript',
    'object',
    'ol',
    'optgroup',
    'option',
    'p',
    'param',
    'pre',
    'q',
    'samp',
    'script',
    'select',
    'small',
    'span',
    'strong',
    'style',
    'sub',
    'sup',
    'table',
    'tbody',
    'td',
    'textarea',
    'tfoot',
    'th',
    'thead',
    'title',
    'tr',
    'tt',
    'ul',
    'var',
}

# From the XHTML 1.0 Transitional DTD
# <http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd>
# (difference to transitional only)
XHTML_1_0_TRANSITIONAL = XHTML_1_0_STRICT | {
    'applet',
    'basefont',
    'center',
    'dir',
    'font',
    'iframe',
    'isindex',
    'menu',
    'noframes',
    's',
    'strike',
    'u',
}

# HTML 5 Elements (difference from XHTML 1.0 Transitional), see
# <https://www.w3.org/TR/html5-diff/#language> and
# <https://html.spec.whatwg.org/>
HTML_5 = XHTML_1_0_TRANSITIONAL - {
    'acronym',
    'applet',
    'basefont',
    'big',
    'center',
    'dir',
    'font',
    'isindex',
    'strike',
    'tt',
} | {
    'article',
    'aside',
    'audio',
    'bdi',
    'canvas',
    'datalist',
    'details',
    'dialog',
    'embed',
    'figcaption',
    'figure',
    'footer',
    'header',
    'keygen',
    'main',
    'mark',
    'meter',
    'nav',
    'output',
    'picture',
    'progress',
    'rp',
    'rt',
    'ruby',
    'section',
    'slot',
    'source',
    'summary',
    'template',
    'time',
    'track',
    'video',
    'wbr',
}

XHTML_1_0_EMPTY_ELEMENTS = {
    'area',
    'base',
    'basefont',
    'br',
    'col',
    'hr',
    'iframe',
    'img',
    'input',
    'isindex',
    'link',
    'meta',
    'param',
}

HTML_5_EMPTY_ELEMENTS = XHTML_1_0_EMPTY_ELEMENTS - {
    'basefont',
    'iframe',
    'isindex',
} | {
    'embed',
    'keygen',
    'source',
    'track',
    'wbr',
}


#
# Class definitions
#


class XmlElement():

    """Callable XML element"""

    fs_empty_element = \
        '<{start_tag}{attributes}/>'
    fs_generic_element = \
        '<{start_tag}{attributes}>{content}</{end_tag}>'
    # Empty set: do not restrict element names
    restrict_elements_to = set()
    name_translations = (
        translate.remove_trailing_underscores,
        translate.underscores_to_dashes,
    )

    def __init__(self, tag_name):
        """Set tag name"""
        tag_name = self.translate_name(tag_name)
        #
        if self.restrict_elements_to and \
                tag_name not in self.restrict_elements_to:
            raise ValueError('Unsupported element name {0!r}'.format(
                tag_name))
        #
        self.tag_name = tag_name

    @classmethod
    def translate_name(cls, name):
        """Apply all functions in the name_translations sequence to name
        and return the result
        """
        for translation_function in cls.name_translations:
            name = translation_function(name)
        #
        return name

    @classmethod
    def single_attribute(cls, attribute_name, attribute_value):
        """Make an XML attribute from the given attr_name, attr_value pair"""
        if attribute_value is None or attribute_value is False:
            return constants.EMPTY
        #
        attribute_name = cls.translate_name(attribute_name)
        #
        if attribute_value is True:
            attribute_value = attribute_name
        #
        return ' {0}={1}'.format(
            attribute_name,
            xml.sax.saxutils.quoteattr(str(attribute_value)))

    @classmethod
    def attributes_string(cls, attributes):
        """Make a single string from the 'attributes' (name, value) list.
        Attributes with None values are ignored.
        """
        tag_attributes = [
            cls.single_attribute(attribute_name, attribute_value)
            for (attribute_name, attribute_value) in attributes
            if attribute_value is not None]
        return constants.EMPTY.join(tag_attributes)

    def _output(self,
                content_fragments,
                attributes,
                compact_empty=True,
                start_tag_override=None):
        """Return the element as a string containing an XML subtree"""
        start_tag = start_tag_override or self.tag_name
        content = constants.EMPTY.join(content_fragments)
        if compact_empty and not content:
            fs_element = self.fs_empty_element
        else:
            fs_element = self.fs_generic_element
        #
        return fs_element.format(
            start_tag=start_tag,
            attributes=self.attributes_string(attributes.items()),
            content=content,
            end_tag=self.tag_name)

    def __call__(self, *content_fragments, **attributes):
        """Return an element generated from the given parameters"""
        return self._output(content_fragments,
                            attributes,
                            compact_empty=True)


class XmlBasedHtmlElement(XmlElement):

    """Callable HTML element (base class)"""

    fs_empty_element = \
        '<{start_tag}{attributes} />'
    # Always lowercase tag names
    name_translations = (
        translate.remove_trailing_underscores,
        translate.underscores_to_dashes,
        str.lower,
    )

    # Definitions for HTML elements only
    empty_elements = XHTML_1_0_EMPTY_ELEMENTS
    attribute_duplications = {}

    def __init__(self,
                 tag_name):
        """Set tag name"""
        super(XmlBasedHtmlElement, self).__init__(tag_name)
        self.is_empty = self.tag_name in self.empty_elements

    def __call__(self, *content_fragments, **attributes):
        """Return a new tag from the given parameters

        Special attributes:
        __class__   -> HTML attribute 'class' because 'class'
                       is a keyword in Python
        __classes__ -> Sequence of classes for the element
        """
        #
        # If a classes sequence was given using the parameter __classes__,
        # construct a new class attribute
        # with all given class names separated by blanks
        classes_set = set()
        try:
            classes_set.add(attributes.pop('__class__'))
        except KeyError:
            ...
        #
        try:
            classes_set.update(set(attributes.pop('__classes__')))
        except KeyError:
            ...
        #
        if classes_set:
            attributes['class'] = constants.BLANK.join(sorted(classes_set))
        #
        # Duplicate certain elements as others
        # (xml:lang= with the same value as lang= in XHTML)
        for (source, target) in self.attribute_duplications.items():
            try:
                attributes[target] = attributes[source]
            except KeyError:
                ...
            #
        #
        if self.is_empty:
            return self._output(tuple(),
                                attributes,
                                compact_empty=True)
        #
        return self._output(content_fragments,
                            attributes,
                            compact_empty=False)


class HtmlElement(XmlBasedHtmlElement):

    """Callable HTML (5) element"""

    fs_empty_element = ('<{start_tag}{attributes}>')
    restrict_elements_to = HTML_5
    empty_elements = HTML_5_EMPTY_ELEMENTS

    @classmethod
    def single_attribute(cls, attribute_name, attribute_value):
        """Make an XML attribute from the given attr_name, attr_value pair"""
        if attribute_value is True:
            return ' {0}'.format(cls.translate_name(attribute_name))
        #
        return XmlElement.single_attribute(
            attribute_name,
            attribute_value)


class XhtmlStrictElement(XmlBasedHtmlElement):

    """Callable XHTML 1.0 Strict element"""

    restrict_elements_to = XHTML_1_0_STRICT
    attribute_duplications = {'lang': 'xml:lang'}


class XhtmlTransitionalElement(XhtmlStrictElement):

    """Callable XHTML 1.0 Transitional element"""

    restrict_elements_to = XHTML_1_0_TRANSITIONAL


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-

"""

smallparts.markup.characters

Markup (HTML, XML) generation – Characters handling

"""


import unicodedata
import xml.sax.saxutils

from smallparts import constants


#
# Constants
#

# Keyword constants for the Defuser class

REMOVE_INVALID = 0
REMOVE_RESTRICTED = 1
REMOVE_DISCOURAGED = 2


#
# Classes
#


class Defuser():

    """Used to defuse a provided text:
    remove invalid and/or restricted and/or discouraged codepoints
    from the text and (XML-)escape the characters < > &
    (the latter is delegated to xml.sax.saxutils.escape()).
    """

    supported_xml_versions = (constants.XML_1_0,
                              constants.XML_1_1)
    supported_removals = (REMOVE_INVALID,
                          REMOVE_RESTRICTED,
                          REMOVE_DISCOURAGED)
    invalid_in_xml_1_1 = \
        (0x0,) + tuple(range(0xd800, 0xe000)) + (0xfffe, 0xffff)
    restricted_c0_controls = \
        tuple(range(0x1, 0x9)) + (0xb, 0xc) + tuple(range(0xe, 0x20))
    restricted_in_xml_1_0 = \
        tuple(range(0x7f, 0x85)) + tuple(range(0x86, 0xa0))

    codepoints = (
        {constants.XML_1_0: invalid_in_xml_1_1 + restricted_c0_controls,
         constants.XML_1_1: invalid_in_xml_1_1},
        {constants.XML_1_0: restricted_in_xml_1_0,
         constants.XML_1_1: restricted_in_xml_1_0 + restricted_c0_controls},
        tuple(range(0xfdd0, 0xfde0)) + (
            0x1fffe, 0x1ffff, 0x2fffe, 0x2ffff, 0x3fffe, 0x3ffff,
            0x4fffe, 0x4ffff, 0x5fffe, 0x5ffff, 0x6fffe, 0x6ffff,
            0x7fffe, 0x7ffff, 0x8fffe, 0x8ffff, 0x9fffe, 0x9ffff,
            0xafffe, 0xaffff, 0xbfffe, 0xbffff, 0xcfffe, 0xcffff,
            0xdfffe, 0xdffff, 0xefffe, 0xeffff, 0xffffe, 0xfffff,
            0x10fffe, 0x10ffff))

    def __init__(self,
                 xml_version=constants.XML_1_0,
                 remove=REMOVE_INVALID):
        """Build the translation dict"""
        if xml_version not in self.supported_xml_versions:
            raise ValueError(
                'xml_version must be one of {0!r}!'.format(
                    self.supported_xml_versions))
        #
        if remove not in self.supported_removals:
            if remove is None:
                raise ValueError(
                    'Please use the {0!r} class static method'
                    ' .escape() if you do not want to remove'
                    ' codepoints at all'.format(self.__class__.__name__))
            #
            raise ValueError(
                'remove must be one of {0!r}!'.format(
                    self.supported_removals))
        #
        self.__removable_codepoints = dict.fromkeys(
            self.codepoints[REMOVE_INVALID][xml_version])
        if remove >= REMOVE_RESTRICTED:
            self.__removable_codepoints.update(
                dict.fromkeys(
                    self.codepoints[REMOVE_RESTRICTED][xml_version]))
        #
        if remove >= REMOVE_DISCOURAGED:
            self.__removable_codepoints.update(
                dict.fromkeys(
                    self.codepoints[REMOVE_DISCOURAGED]))
        #

    @staticmethod
    def escape(source_text):
        """Wrap xml.sax.saxutils.escape()"""
        return xml.sax.saxutils.escape(source_text)

    def remove_codepoints(self, source_text):
        """Remove codepoints as specified at instantiation time"""
        return source_text.translate(self.__removable_codepoints)

    def defuse(self, source_text):
        """Remove codepoints as specified at instantiation time,
        and escape the remaining characters
        """
        return self.escape(self.remove_codepoints(source_text))


#
# Functions
#


def encode_to_charrefs(source_text):
    """Replace non-ascii characters by numeric entities"""
    ascii_bytes = source_text.encode('ascii',
                                     errors='xmlcharrefreplace')
    return ascii_bytes.decode()


def entity(reference):
    """Return a numeric (&#reference;) or symbolic: (&reference;) entity,
    depending on the reference's type
    """
    try:
        return '&#{0:d};'.format(reference)
    except ValueError:
        return '&{0};'.format(reference)
    #


def charref_from_name(unicode_character_name):
    """Return the numeric (&#reference;) entity
    for the given unicode character name
    """
    return entity(ord(unicodedata.lookup(unicode_character_name)))


def translate_to_charrefs(characters_sequence, source_text):
    """Return source_text with all characters from the
    characters sequence translated to their respective charrefs.
    """
    charrefs = {}
    for character in characters_sequence:
        codepoint = ord(character)
        charrefs[codepoint] = entity(codepoint)
    #
    return source_text.translate(charrefs)


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-

"""

smallparts.l10n.time_indications

Time indication translations

"""


from smallparts.l10n import languages


SECONDS = 'seconds'
MINUTES = 'minutes'
HOURS = 'hours'
DAYS = 'days'
WEEKS = 'weeks'
MONTHS = 'months'
YEARS = 'years'

SUPPORTED_UNITS = (SECONDS, MINUTES, HOURS, DAYS,
                   WEEKS, MONTHS, YEARS)


# Singular and plural forms for time units
NUMBER_CATEGORIES = {
    languages.EN: {
        SECONDS: ('second', SECONDS),
        MINUTES: ('minute', MINUTES),
        HOURS: ('hour', HOURS),
        DAYS: ('day', DAYS),
        WEEKS: ('week', WEEKS),
        MONTHS: ('month', MONTHS),
        YEARS: ('year', YEARS)
    },
    languages.DE: {
        SECONDS: ('Sekunde', 'Sekunden'),
        MINUTES: ('Minute', 'Minuten'),
        HOURS: ('Stunde', 'Stunden'),
        DAYS: ('Tag', 'Tage'),
        WEEKS: ('Woche', 'Wochen'),
        MONTHS: ('Monat', 'Monate'),
        YEARS: ('Jahr', 'Jahre')
    },
    languages.FR: {
        SECONDS: ('seconde', 'secondes'),
        MINUTES: ('minute', 'Minutes'),
        HOURS: ('heure', 'heures'),
        DAYS: ('jour', 'jours'),
        WEEKS: ('semaine', 'semaines'),
        MONTHS: ('mois', 'mois'),
        YEARS: ('an', 'ans')
    }
}


#
# Functions
#


def pretty_print_component(**kwargs):
    """Return the time component,
    pretty printed in singular or plural form
    in the language selected by lang=...
    """
    lang = kwargs.get('lang') or languages.DEFAULT
    try:
        number_category = NUMBER_CATEGORIES[lang]
    except KeyError:
        raise ValueError(languages.missing_translation(lang))
    #
    for unit in SUPPORTED_UNITS:
        amount = kwargs.get(unit)
        if amount is not None:
            singular, plural = number_category[unit]
            if amount == 1:
                matching_form = singular
            else:
                matching_form = plural
            #
            break
        #
    else:
        raise ValueError('Please specify an amount of time as keyword'
                         ' argument {0!r}!'.format(SUPPORTED_UNITS))
    #
    return '{0} {1}'.format(amount, matching_form)


# vim:fileencoding=utf-8 autoindent ts=4 sw=4 sts=4 expandtab:



# -*- coding: utf-8 -*-

"""

smallparts.markup.generators

Markup (HTML, XML) generation

"""


from smallparts import constants

from smallparts.markup import elements
from smallparts.namespaces import Namespace
from smallparts.text import join


#
# Constants
#

HTML_5 = 'HTML 5'
XHTML_1_0_STRICT = 'XHTML 1.0 Strict'
XHTML_1_0_TRANSITIONAL = 'XHTML 1.0 Transitional'
XML = 'XML'

SUPPORTED_DIALECTS = {
    HTML_5: Namespace(
        doctype='<!DOCTYPE html>',
        factory=elements.HtmlElement,
        xmlns=None),
    XHTML_1_0_STRICT: Namespace(
        doctype='<!DOCTYPE html PUBLIC'
        ' "-//W3C//DTD XHTML 1.0 Strict//EN"'
        ' "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">',
        factory=elements.XhtmlStrictElement,
        xmlns='http://www.w3.org/1999/xhtml'),
    XHTML_1_0_TRANSITIONAL: Namespace(
        doctype='<!DOCTYPE html PUBLIC'
        ' "-//W3C//DTD XHTML 1.0 Transitional//EN"'
        ' "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">',
        factory=elements.XhtmlTransitionalElement,
        xmlns='http://www.w3.org/1999/xhtml'),
    XML: Namespace(factory=elements.XmlElement),
}


#
# Class definitions
#


# pylint: disable= too-few-public-methods ; Not required for the cache


class ElementsCache():

    """Cache for element factories"""

    accessible_attributes = ('__class__',
                             '_dialect',
                             '_factory',
                             '__repr__')
    _cached_elements = {
        HTML_5: {},
        XHTML_1_0_STRICT: {},
        XHTML_1_0_TRANSITIONAL: {},
        XML: {}}

    def __init__(self, dialect):
        """Initialize the cache"""
        if dialect not in type(self)._cached_elements:
            raise ValueError('Unsupported dialect.')
        #
        self._dialect = dialect
        self._factory = SUPPORTED_DIALECTS[self._dialect].factory

    def __repr__(self):
        """Textual representation"""
        return '{0}(dialect={1!r})'.format(
            self.__class__.__name__, self._dialect)

    def __getattribute__(self, name):
        """Return an existing cache member
        or create a new member
        """
        if name in type(self).accessible_attributes:
            return object.__getattribute__(self, name)
        #
        name = self._factory.translate_name(name)
        try:
            return type(self)._cached_elements[self._dialect][name]
        except KeyError:
            return type(self)._cached_elements[self._dialect].setdefault(
                name,
                self._factory(name))


#
# Functions
#


def css_property(property_name, property_value):
    """Generate a CSS property:
    property_name: property_value;
    """
    return '{0}: {1};'.format(property_name, property_value)


def css_important_property(property_name, property_value):
    """Generate an 'important' CSS property:
    property_name: property_value !important;
    """
    return css_property(property_name,
                        '{0} !important'.format(property_value))


def html_document(dialect=HTML_5,
                  lang='en',
                  title='Untitled page',
                  head='',
                  body=''):
    """Generate an HTML document"""
    if dialect == XML:
        raise ValueError('This function can only generate HTML documents.')
    #
    try:
        html_dialect = SUPPORTED_DIALECTS[dialect]
    except KeyError:
        raise ValueError(
            'Unsupported HTML dialect.'
            ' Please specify one of {0}!'.format(
                constants.COMMA_BLANK.join(
                    repr(single_dialect) for single_dialect in
                    SUPPORTED_DIALECTS if 'HTML' in single_dialect)))
    #
    element = ElementsCache(dialect)
    head_fragments = ['']
    if dialect == HTML_5 and '<meta charset' not in head.lower():
        head_fragments.append(element.meta(charset=constants.UTF_8))
    #
    if '<title' not in head.lower():
        head_fragments.append(element.title(title))
    #
    head = head.strip()
    if head:
        head_fragments.append(head)
    #
    head_fragments.append('')
    body = body.strip()
    if body:
        body = '\n{0}\n'.format(body)
    return join.by_newlines(
        html_dialect.doctype,
        element.html(
            join.by_newlines(
                '',
                element.head(constants.NEWLINE.join(head_fragments)),
                element.body(body),
                ''),
            xmlns=html_dialect.xmlns,
            lang=lang))


def js_function_call(function_name, arguments):
    """Generate JavaScript code:
    function_name(*arguments)
    """
    return '{0}({1})'.format(
        function_name,
        constants.COMMA_BLANK.join(
            "{0!r}".format(single_arg)
            for single_arg in arguments))


def js_return(function_name, *arguments):
    """Generate JavaScript code:
    return function_name(*arguments);
    """
    return 'return {0};'.format(
        js_function_call(function_name, arguments))


def wrap_cdata(character_data):
    """Wrap character_data in a CDATA section.
    If necessary use multiple CDATA sections as suggested in
    <https://en.wikipedia.org/wiki/CDATA#Nesting>
    """
    return join.directly(
        '<![CDATA[',
        character_data.replace(']]>', ']]]]><![CDATA[>'),
        ']]>')


def xml_declaration(version=constants.XML_1_0,
                    encoding=constants.UTF_8,
                    standalone=None):
    """Return an XML declaration.
    Omit the 'standalone' attribute if not specified.
    """
    if standalone is not None:
        if standalone:
            standalone = constants.YES
        else:
            standalone = constants.NO
        #
    #
    return '<?xml{0} ?>'.format(
        elements.XmlElement.attributes_string(
            dict(version=version,
                 encoding=encoding,
                 standalone=standalone).items()))


def xml_document(content,
                 version=constants.XML_1_0,
                 encoding=constants.UTF_8,
                 standalone=None):
    """Return a full XML document.
    Strip trailing whitespace from the content.
    """
    return join.by_newlines(
        xml_declaration(version=version,
                        encoding=encoding,
                        standalone=standalone),
        content.rstrip())


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:
